/*=========================================================================
s
  Program:   Flow Topology
  Module:    $RCSfile: vtkFlowTopo.cxx,v $

     This software is distributed WITHOUT ANY WARRANTY; without even
     the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
     PURPOSE.  See the above copyright notice for more information.

=========================================================================*/
#include "vtkFlowTopo.h"

#include "vtkCell.h"
#include "vtkIdList.h"
#include "vtkInformation.h"
#include "vtkInformationVector.h"
#include "vtkObjectFactory.h"
#include "vtkPointData.h"
#include "vtkCellArray.h"
#include "vtkPolyData.h"
#include "vtkCellType.h"

//#include "linalg.h"

vtkCxxRevisionMacro(vtkFlowTopo, "$Revision: 0.01 $");
vtkStandardNewMacro(vtkFlowTopo);

const bool skipBoundaryCells = false;

vtkFlowTopo::vtkFlowTopo()
{ // nothing to be modified inside this method

  GlyphRadius = 20.0;

  this->SetInputArrayToProcess(0, 0, 0, vtkDataObject::FIELD_ASSOCIATION_POINTS,
                               vtkDataSetAttributes::VECTORS);
}

vtkFlowTopo::~vtkFlowTopo()
{ // nothing to be modified inside this method
}

void vtkFlowTopo::generateCross(double posX, double posY, double posZ,
				vtkPoints *outputPoints,
				vtkCellArray *outputLines,
				int &numPoints)
{ // nothing to be modified inside this method
  // cross of three orthogonal lines
  // -> each line is represented by a VTK cell that contains 2 points

  for (int line=0; line<3; line++) {

    outputLines->InsertNextCell(2);
    for (int v=0; v<2; v++) {

      double pnt[3] = { posX, posY, posZ };
      if (v == 0) pnt[line] -= GlyphRadius / 2.0;
      else pnt[line] += GlyphRadius / 2.0;
    
      // store point
      outputPoints->InsertNextPoint(pnt);

      // insert point ID into cell (connectivity info)
      outputLines->InsertCellPoint(numPoints++);
    }
  }
}

bool boundaryCell(vtkUnstructuredGrid *grid, int cell)
{ // nothing to be modified inside this method

  bool onBoundary = false;
  for (int face=0; face<4; face++) {
    
    // get face points
    vtkIdList *ptIds = vtkIdList::New();
    ptIds->SetNumberOfIds(3);
    int pntCnt = 0;
    for (int p=0; p<4; p++) {
      if (p != face) {
	ptIds->SetId(pntCnt++, grid->GetCell(cell)->GetPointId(p));
      }
    }
    
    // lookup cell that shares the current face
    vtkIdList *cellIds = vtkIdList::New();
    cellIds->SetNumberOfIds(1);
    grid->GetCellNeighbors(cell, ptIds, cellIds);
    
    onBoundary = cellIds->GetNumberOfIds() < 1;
    
    ptIds->Delete();
    cellIds->Delete();
    
    if (onBoundary) {
      // boundary cell
      break;;
    }
  }

  return onBoundary;
}

#if 0
// code snippets

// go over nodes and get position and vector value on nodes
{
  // get node IDs of cell
  vtkIdList *cellNodeIds = grid->GetCell(cell)->GetPointIds();

  for (int node=0; node<4; node++) {
    // get node position
    float *positions = 
      ((vtkFloatArray *) grid->GetPoints()->GetData())->GetPointer(0);
    float *pos = &positions[cellNodeIds->GetId(node)*3];

    // get vector at node
    float vecW[3];
    inVectors->GetTupleValue(cellNodeIds->GetId(node), vecW);
  }
}
#endif

void vtkFlowTopo::computeIt(vtkUnstructuredGrid *grid,
			    vtkFloatArray *inVectors,
			    vtkPolyData *polyDat)
{ // add (call) your implementation at the '...'

  int geomType;
  vtkPolyData *vtkPolyDat = polyDat;
  vtkPoints *outputPoints;
  vtkCellArray *outputLines;
  int numPoints = 0;
  int numLines = 0;
  int numCritPoints = 0;

  outputPoints = vtkPoints::New();
  outputLines = vtkCellArray::New();

  // go over unstructured cells
  for (int c=0; c<grid->GetNumberOfCells(); c++) {

    if (grid->GetCellType(c) != VTK_TETRA) {
      printf("skipping cell %d because cell type not supported\n", c);
    }

    // skip tetrahedra at boundary
    if (skipBoundaryCells) {
      if (boundaryCell(grid, c)) continue;
    }

    // early rejection of cells without critical points
    // also evaluates which edges contain a zero of the first vector component
    // ...

    // find critical point in cell
    float point[3];
    // ...
#if 1 // for single glyph test
    point[0] = 0.0;
    point[1] = 0.0;
    point[2] = 0.0;
#endif

    numCritPoints++;

    // generate glyph
    generateCross(point[0], point[1], point[2], outputPoints, outputLines, numPoints);

#if 1 // for single glyph test
    break;
#endif
  }

  printf("%d critical points\n", numCritPoints);

  // set geometry
  vtkPolyDat->SetPoints(outputPoints);
  vtkPolyDat->SetLines(outputLines);
 
  // reference is counted, we can delete
  outputPoints->Delete();
  outputLines->Delete();
}

int vtkFlowTopo::RequestData(vtkInformation *vtkNotUsed(request),
			     vtkInformationVector **inputVector,
			     vtkInformationVector *outputVector)
{ // nothing to be modified inside this method

  // get the info objects
  vtkInformation *inInfo = inputVector[0]->GetInformationObject(0);
  vtkInformation *outInfo = outputVector->GetInformationObject(0);

  // get the input and ouptut
  vtkUnstructuredGrid *input =
    vtkUnstructuredGrid::SafeDownCast(inInfo->Get(vtkDataObject::DATA_OBJECT()));
  vtkPolyData *output =
    vtkPolyData::SafeDownCast(outInfo->Get(vtkDataObject::DATA_OBJECT()));

  vtkIdType numCells = input->GetNumberOfCells();

  // only supporting Float!
  vtkFloatArray *inVectors = (vtkFloatArray *) this->GetInputArrayToProcess(0, inputVector);

  // Initialize
  vtkDebugMacro(<<"Computing flow topology (critical points)");

  // Check input
  if (numCells < 1) { 
    printf("no cells\n");
    vtkErrorMacro("No cells");
    return 1;
  }

  if (!inVectors) {
    printf("no vector data\n");
    vtkDebugMacro(<<"No vector data");
    return 1;
  }
  
  // compute
  computeIt(input, inVectors, output);
 
  return 1;
}

int vtkFlowTopo::FillInputPortInformation(int, vtkInformation *info)
{ // nothing to be modified inside this method
  info->Set(vtkAlgorithm::INPUT_REQUIRED_DATA_TYPE(), "vtkDataSet");
  return 1;
}

void vtkFlowTopo::PrintSelf(ostream& os, vtkIndent indent)
{ // nothing to be modified inside this method
  this->Superclass::PrintSelf(os,indent);
}
